@@ -2,13 +2,15 @@
 
 use std::fmt::{Display, Formatter};
 
-use dioxus_core::exports::bumpalo;
-use dioxus_core::exports::bumpalo::Bump;
 use dioxus_core::IntoVNode;
 use dioxus_core::*;
 
+fn app(_cx: Scope) -> Element {
+    None
+}
+
 pub struct SsrRenderer {
-    inner: bumpalo::Bump,
+    vdom: VirtualDom,
     cfg: SsrConfig,
 }
 
@@ -16,21 +18,14 @@ impl SsrRenderer {
     pub fn new(cfg: impl FnOnce(SsrConfig) -> SsrConfig) -> Self {
         Self {
             cfg: cfg(SsrConfig::default()),
-            inner: bumpalo::Bump::new(),
+            vdom: VirtualDom::new(app),
         }
     }
 
     pub fn render_lazy<'a>(&'a mut self, f: LazyNodes<'a, '_>) -> String {
-        let bump = &mut self.inner as *mut _;
-        let s = self.render_inner(f);
+        let scope = self.vdom.base_scope();
+        let factory = NodeFactory::new(scope);
 
-        // reuse the bump's memory
-        unsafe { (&mut *bump as &mut bumpalo::Bump).reset() };
-        s
-    }
-
-    fn render_inner<'a>(&'a self, f: LazyNodes<'a, '_>) -> String {
-        let factory = NodeFactory::new(&self.inner);
         let root = f.into_vnode(factory);
         format!(
             "{:}",
@@ -43,9 +38,10 @@ impl SsrRenderer {
     }
 }
 
+#[allow(clippy::needless_lifetimes)]
 pub fn render_lazy<'a>(f: LazyNodes<'a, '_>) -> String {
-    let bump = bumpalo::Bump::new();
-    let borrowed = &bump;
+    let vdom = VirtualDom::new(app);
+    let scope: *const ScopeState = vdom.base_scope();
 
     // Safety
     //
@@ -55,11 +51,11 @@ pub fn render_lazy<'a>(f: LazyNodes<'a, '_>) -> String {
     // When LazyNodes are provided, they are FnOnce, but do not come with a allocator selected to borrow from. The <'a>
     // lifetime is therefore longer than the lifetime of the allocator which doesn't exist... yet.
     //
-    // Therefore, we cast our local bump alloactor into right lifetime. This is okay because our usage of the bump arena
-    // is *definitely* shorter than the <'a> lifetime, and we return *owned* data - not borrowed data.
+    // Therefore, we cast our local bump allocator to the right lifetime. This is okay because our usage of the bump
+    // arena is *definitely* shorter than the <'a> lifetime, and we return *owned* data - not borrowed data.
+    let scope = unsafe { &*scope };
 
-    let _b = unsafe { std::mem::transmute::<&Bump, &'a Bump>(borrowed) };
-    let root = f.into_vnode(NodeFactory::new(_b));
+    let root = f.into_vnode(NodeFactory::new(scope));
 
     format!(
         "{:}",
@@ -75,6 +71,13 @@ pub fn render_vdom(dom: &VirtualDom) -> String {
     format!("{:}", TextRenderer::from_vdom(dom, SsrConfig::default()))
 }
 
+pub fn pre_render_vdom(dom: &VirtualDom) -> String {
+    format!(
+        "{:}",
+        TextRenderer::from_vdom(dom, SsrConfig::default().pre_render(true))
+    )
+}
+
 pub fn render_vdom_cfg(dom: &VirtualDom, cfg: impl FnOnce(SsrConfig) -> SsrConfig) -> String {
     format!(
         "{:}",
@@ -119,7 +122,8 @@ pub struct TextRenderer<'a, 'b> {
 
 impl Display for TextRenderer<'_, '_> {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
-        self.html_render(self.root, f, 0)
+        let mut last_node_was_text = false;
+        self.html_render(self.root, f, 0, &mut last_node_was_text)
     }
 }
 
@@ -132,26 +136,42 @@ impl<'a> TextRenderer<'a, '_> {
         }
     }
 
-    fn html_render(&self, node: &VNode, f: &mut std::fmt::Formatter, il: u16) -> std::fmt::Result {
+    fn html_render(
+        &self,
+        node: &VNode,
+        f: &mut std::fmt::Formatter,
+        il: u16,
+        last_node_was_text: &mut bool,
+    ) -> std::fmt::Result {
         match &node {
             VNode::Text(text) => {
+                if *last_node_was_text {
+                    write!(f, "<!--spacer-->")?;
+                }
+
                 if self.cfg.indent {
                     for _ in 0..il {
                         write!(f, "    ")?;
                     }
                 }
+
+                *last_node_was_text = true;
+
                 write!(f, "{}", text.text)?
             }
             VNode::Placeholder(_anchor) => {
-                //
+                *last_node_was_text = false;
+
                 if self.cfg.indent {
                     for _ in 0..il {
                         write!(f, "    ")?;
                     }
                 }
-                write!(f, "<!-- -->")?;
+                write!(f, "<!--placeholder-->")?;
             }
             VNode::Element(el) => {
+                *last_node_was_text = false;
+
                 if self.cfg.indent {
                     for _ in 0..il {
                         write!(f, "    ")?;
@@ -167,6 +187,36 @@ impl<'a> TextRenderer<'a, '_> {
                     match attr.namespace {
                         None => match attr.name {
                             "dangerous_inner_html" => inner_html = Some(attr.value),
+                            "allowfullscreen"
+                            | "allowpaymentrequest"
+                            | "async"
+                            | "autofocus"
+                            | "autoplay"
+                            | "checked"
+                            | "controls"
+                            | "default"
+                            | "defer"
+                            | "disabled"
+                            | "formnovalidate"
+                            | "hidden"
+                            | "ismap"
+                            | "itemscope"
+                            | "loop"
+                            | "multiple"
+                            | "muted"
+                            | "nomodule"
+                            | "novalidate"
+                            | "open"
+                            | "playsinline"
+                            | "readonly"
+                            | "required"
+                            | "reversed"
+                            | "selected"
+                            | "truespeed" => {
+                                if attr.value != "false" {
+                                    write!(f, " {}=\"{}\"", attr.name, attr.value)?
+                                }
+                            }
                             _ => write!(f, " {}=\"{}\"", attr.name, attr.value)?,
                         },
 
@@ -189,18 +239,6 @@ impl<'a> TextRenderer<'a, '_> {
                     }
                 }
 
-                // we write the element's id as a data attribute
-                //
-                // when the page is loaded, the `querySelectorAll` will be used to collect all the nodes, and then add
-                // them interpreter's stack
-                if let (true, Some(id)) = (self.cfg.pre_render, node.try_mounted_id()) {
-                    write!(f, " dioxus-id=\"{}\"", id)?;
-
-                    for _listener in el.listeners {
-                        // todo: write the listeners
-                    }
-                }
-
                 match self.cfg.newline {
                     true => writeln!(f, ">")?,
                     false => write!(f, ">")?,
@@ -209,8 +247,9 @@ impl<'a> TextRenderer<'a, '_> {
                 if let Some(inner_html) = inner_html {
                     write!(f, "{}", inner_html)?;
                 } else {
+                    let mut last_node_was_text = false;
                     for child in el.children {
-                        self.html_render(child, f, il + 1)?;
+                        self.html_render(child, f, il + 1, &mut last_node_was_text)?;
                     }
                 }
 
@@ -230,7 +269,7 @@ impl<'a> TextRenderer<'a, '_> {
             }
             VNode::Fragment(frag) => {
                 for child in frag.children {
-                    self.html_render(child, f, il + 1)?;
+                    self.html_render(child, f, il + 1, last_node_was_text)?;
                 }
             }
             VNode::Component(vcomp) => {
@@ -238,7 +277,7 @@ impl<'a> TextRenderer<'a, '_> {
 
                 if let (Some(vdom), false) = (self.vdom, self.cfg.skip_components) {
                     let new_node = vdom.get_scope(idx).unwrap().root_node();
-                    self.html_render(new_node, f, il + 1)?;
+                    self.html_render(new_node, f, il + 1, last_node_was_text)?;
                 } else {
                 }
             }
@@ -282,154 +321,3 @@ impl SsrConfig {
         self
     }
 }
-
-#[cfg(test)]
-mod tests {
-    use super::*;
-
-    use dioxus_core::prelude::*;
-    use dioxus_core_macro::*;
-    use dioxus_html as dioxus_elements;
-
-    static SIMPLE_APP: Component = |cx| {
-        cx.render(rsx!(div {
-            "hello world!"
-        }))
-    };
-
-    static SLIGHTLY_MORE_COMPLEX: Component = |cx| {
-        cx.render(rsx! {
-            div { title: "About W3Schools",
-                (0..20).map(|f| rsx!{
-                    div {
-                        title: "About W3Schools",
-                        style: "color:blue;text-align:center",
-                        class: "About W3Schools",
-                        p {
-                            title: "About W3Schools",
-                            "Hello world!: {f}"
-                        }
-                    }
-                })
-            }
-        })
-    };
-
-    static NESTED_APP: Component = |cx| {
-        cx.render(rsx!(
-            div {
-                SIMPLE_APP {}
-            }
-        ))
-    };
-    static FRAGMENT_APP: Component = |cx| {
-        cx.render(rsx!(
-            div { "f1" }
-            div { "f2" }
-            div { "f3" }
-            div { "f4" }
-        ))
-    };
-
-    #[test]
-    fn to_string_works() {
-        let mut dom = VirtualDom::new(SIMPLE_APP);
-        dom.rebuild();
-        dbg!(render_vdom(&dom));
-    }
-
-    #[test]
-    fn hydration() {
-        let mut dom = VirtualDom::new(NESTED_APP);
-        dom.rebuild();
-        dbg!(render_vdom_cfg(&dom, |c| c.pre_render(true)));
-    }
-
-    #[test]
-    fn nested() {
-        let mut dom = VirtualDom::new(NESTED_APP);
-        dom.rebuild();
-        dbg!(render_vdom(&dom));
-    }
-
-    #[test]
-    fn fragment_app() {
-        let mut dom = VirtualDom::new(FRAGMENT_APP);
-        dom.rebuild();
-        dbg!(render_vdom(&dom));
-    }
-
-    #[test]
-    fn write_to_file() {
-        use std::fs::File;
-        use std::io::Write;
-
-        let mut file = File::create("index.html").unwrap();
-
-        let mut dom = VirtualDom::new(SLIGHTLY_MORE_COMPLEX);
-        dom.rebuild();
-
-        file.write_fmt(format_args!(
-            "{}",
-            TextRenderer::from_vdom(&dom, SsrConfig::default())
-        ))
-        .unwrap();
-    }
-
-    #[test]
-    fn styles() {
-        static STLYE_APP: Component = |cx| {
-            cx.render(rsx! {
-                div { color: "blue", font_size: "46px"  }
-            })
-        };
-
-        let mut dom = VirtualDom::new(STLYE_APP);
-        dom.rebuild();
-        dbg!(render_vdom(&dom));
-    }
-
-    #[test]
-    fn lazy() {
-        let p1 = SsrRenderer::new(|c| c).render_lazy(rsx! {
-            div { "ello"  }
-        });
-
-        let p2 = render_lazy(rsx! {
-            div {
-                "ello"
-            }
-        });
-        assert_eq!(p1, p2);
-    }
-
-    #[test]
-    fn big_lazy() {
-        let s = render_lazy(rsx! {
-            div {
-                div {
-                    div {
-                        h1 { "ello world" }
-                        h1 { "ello world" }
-                        h1 { "ello world" }
-                        h1 { "ello world" }
-                        h1 { "ello world" }
-                    }
-                }
-            }
-        });
-
-        dbg!(s);
-    }
-
-    #[test]
-    fn inner_html() {
-        let s = render_lazy(rsx! {
-            div {
-                dangerous_inner_html: "<div> ack </div>"
-            }
-        });
-
-        dbg!(s);
-    }
-}
